
// bouncing ball -- simple hybrid simulation example in C
// (without next-event algorithm for scheduled events)
// Public Domain, Author: PP

#include <stdio.h>
#include <stdbool.h>

double time = 0;                // model time
// simulation parameters
#define ENDTIME  4.0
#define STEPSIZE 0.001          // Warning: limited accuracy of Euler method
#define MINSTEP  1e-6           // used for bisection

//////////////////////////////////////////////////////////////////
// Bouncing ball model definition
#define N 2
double y[N] = { 1, 0 };         // initial state

const double g = 9.81;          // gravity acceleration
void dynamic(double *y, double *yin) // model behavior: y'' = -g
{
    yin[0] = y[1];		// y'
    yin[1] = -g;		// y''
}

// state condition and associated state event
bool cond;			// state condition status
bool eval_state_condition()     // evaluation
{
    return (y[0] >= 0.0);       // height>=0
}

void state_event() {
    if(cond) return;            // TRUE->FALSE only (not optimal solution)
    y[1] = -y[1] * 0.8;         // loss of speed y[1] after collision
    printf("# state event: time=%f\n", time);
    printf("# ========================================\n");
}

//////////////////////////////////////////////////////////////////
// Simulation output (for each step of numerical integration)
void print() {
    printf("%12.8f % 12.10f % 12.10f %d\n", time, y[0], y[1], (int)cond);
}

//////////////////////////////////////////////////////////////////
// Support for state condition change detection algorithm
double ysave[N];                // saved state

void save_state() {
    for(int i=0; i<N; i++)
        ysave[i]=y[i];
}
void restore_state() {
    for(int i=0; i<N; i++)
        y[i]=ysave[i];
}
void init_state_condition()     // initialization
{
    cond = eval_state_condition();
}

//////////////////////////////////////////////////////////////////
// Numerical integration method
// (we use explicit Euler method - suitable for demonstration only)
void continuous_step(double *y, double h)
{
    double yin[N];              // integrators inputs
    dynamic(y,yin);             // evaluate inputs: f(y(t),t)
    for (int i = 0; i < N; i++) // for each integrator
	y[i] += h * yin[i];     // y(t+h) = y(t) + h * f(y(t),t)
    time += h;                  // advance the simulation time
}

//////////////////////////////////////////////////////////////////
// Experiment control
int main()                      // simulation experiment
{
    double h = STEPSIZE;        // continuous simulation step
    init_state_condition();	// state condition initial value
    while (time < ENDTIME) {
        // adjust last step size
        if(time+h*1.01>ENDTIME) h=ENDTIME-time; // do not step over end
        print();
	double last = time;	// mark start of step
	save_state();	        // *** we need saved state for restart
	continuous_step(y,h);
	bool newcond = eval_state_condition();
	if (newcond != cond) {	// state condition change detected
	    if (h <= MINSTEP) {	// can not shorten the step
		cond = newcond;	// new status of condition
                state_event();  // ### do state event
		h = STEPSIZE;	// default step size
	    } else {
		// ### restart step
		time = last;	        // restore time
	        restore_state();        // *** restore state
		h /= 2;		        // shorten next step
		if (h < MINSTEP)
		    h = MINSTEP;        // limit the step size
		printf("# new step size: h=%g\n", h);
	    }
	}
    } // while
    print();
    printf("# end of simulation at time=%f\n", time);
    return 0;
}
